/* ***************************************************************** *
 * Atlantis BBS - CGI Parser                                         *
 * ----------------------------------------------------------------- *
 * Version . . . . . . . . . . . . . . . . . . . . . . . . . . 1.0.6 *
 * ================================================================= *
 * Creation date . . . . . . . . . . . . . . . . . . Wed 12 Aug 1998 *
 * Modification date . . . . . . . . . . . . . . . . Sun 17 Jan 1999 *
 * ================================================================= *
 * Copyright (C) 1997-1998 Free Software Foundation, Inc.            *
 *                                                                   *
 * This program  is  free software;  you can  redistribute it and/or *
 * modify it  under the terms  of the GNU  General Public License as *
 * published by  the Free Software Foundation;  either version 2, or *
 * (at your option) any later version.                               *
 *                                                                   *
 * This program  is distributed in the hope  that it will be useful, *
 * but WITHOUT  ANY WARRANTY;  without even  the implied warranty of *
 * MERCHANTABILITY  or  FITNESS  FOR A  PARTICULAR PURPOSE.  See the *
 * GNU General Public License for more details.                      *
 *                                                                   *
 * You should have received a copy of the GNU General Public License *
 * along  with  this  program;  if not,  write to  the Free Software *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.         *
 *                                                                   *
 * ================================================================= *
 * Authors                                                           *
 *   Sebastien APERGHIS-TRAMONI <saper@esil.univ-mrs.fr>             *
 *                                                                   *
 * If you find any bug, or if you do any improvment, please send us  *
 * a mail (possibly with diff and/or correction).                    *
 *                                                                   *
 * ***************************************************************** */


/* ***************************************************************** *
 * standart includes                                                 *
 * ***************************************************************** */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>


/* ***************************************************************** *
 * CGI includes                                                      *
 * ***************************************************************** */
#include "CGI_parser.h"
#include "log_error.h"


/* ***************************************************************** *
 * defines                                                           *
 * ***************************************************************** */
#define check_pointer(test, action)                                  \
            if( test ) {                                             \
                fprintf(stderr, "%s: new[] failed\n", program);      \
                action;                                              \
            }


/* ***************************************************************** *
 * prototypes                                                        *
 * ***************************************************************** */
char *      get_query       (void);
int         hex2char        (char * hex);
void        plus2space      (char * str);


/* ***************************************************************** *
 * global variables                                                  *
 * ***************************************************************** */
static char program[]   = "CGI Parser";
static char version[]   = "1.0.6";
static char crlf[]      = "\015\012";   // Network end-of-line
static char null[]      = "";



/* ***************************************************************** *
 * expect()                                                          *
 * ------                                                            *
 * description                                                       *
 *   This function allow you to allocate a new array to receive      *
 *   key/value data. You can understand it as "I expect <number>     *
 *   potential values". Returns the array or NULL.                   *
 *                                                                   *
 * ***************************************************************** */
HASH * expect(int number) {
    HASH *  t = NULL;
    
    if(number == 0) 
        return NULL;
    
    t = (HASH *)malloc(sizeof(HASH) * number);
    
    return t;
}


/* ***************************************************************** *
 * purge()                                                           *
 * -----                                                             *
 * description                                                       *
 *   This function purge the given array from memory.                *
 *                                                                   *
 * ***************************************************************** */
void purge(HASH * data) {
    int     i = 0;
    
    if(data == NULL) {
        return;
    }
    
    while(data[i].key != NULL) {
        free(data[i].key);
        free(data[i].value);
        i++;
    }
    
    free(data);
}


/* ***************************************************************** *
 * get_value()                                                       *
 * ---------                                                         *
 * description                                                       *
 *   This function searches for the given keyin the given data[]    *
 *   array, and returns the corresponding value, or NULL if nothing  *
 *   was found.                                                      *
 *                                                                   *
 * ***************************************************************** */
char * get_value(HASH * data, char * key) {
    int     i = 0;
    
    if(data == NULL || key == NULL) {
        return NULL;
    }
    
    for(i=0; data[i].key != NULL; i++) {
        if(strcmp(data[i].key, key) == 0) 
            return data[i].value;
    }
    
    return NULL;
}


/* ***************************************************************** *
 * parse()                                                           *
 * -----                                                             *
 * description                                                       *
 *   This function gets, unescapes and splits the query into         *
 *   key/value tokens, then puts them in the data[] array.           *
 *   Returns the number of items.                                    *
 *                                                                   *
 * ***************************************************************** */
int parse(HASH * data) {
    char *  query = NULL;
    char *  token = NULL;
    int     i = 0;
    
    // Getting back the query. 
    query = get_query();
    
    // If there's a problem, we quit. 
    if(query == NULL) {
        return 0;
    }
    
    // Unescape spaces.
    plus2space(query);
    
    // We use strtok() to read the arguments names and values, which
    // have the following pattern:
    // 
    //       key1=val1&key2=val2
    // 
    // So we make successive calls to strtok(), using "&" and "=" as
    // separator, in alternance. For example, with the query string
    // "foo=bar&pipo=poil&toto=yes", we'll do the following:
    // 
    //   1st call:  token = strtok(query, "=");   // key 'foo'
    //      query: "foo bar&pipo=poil&toto=yes"
    //        token-^
    // 
    //   2nd call:  token = strtok(NULL,  "&");   // value 'bar'
    //      query: "foo bar pipo=poil&toto=yes"
    //            token-^
    // 
    //   3rd call:  token = strtok(NULL,  "=");   // key 'pipo'
    //      query: "foo bar pipo poil&toto=yes"
    //                token-^
    // 
    //   loop to the 2nd call unless token == NULL
    // 
    
    token = strtok(query, "=&");
    
    while(token != NULL) {
        // Reading the key.
        unescape(token);
        data[i].key = new char[strlen(token) + 1];
        check_pointer(data[i].key == NULL, return 0);
        strcpy(data[i].key, token);
        
        token = strtok(NULL, "&");
        
        // Reading the value.
        unescape(token);
        data[i].value = new char[strlen(token) + 1];
        check_pointer(data[i].value == NULL, return 0);
        strcpy(data[i].value, token);
        
        token = strtok(NULL, "=&");
        
        i++;
    }
    
    return i;
}


/* ***************************************************************** *
 * get_query()                                                       *
 * ---------                                                         *
 * description                                                       *
 *   This function gets and returns the sended query for both GET    *
 *   and POST methods.                                               *
 *                                                                   *
 * ***************************************************************** */
char * get_query(void) {
    char *  query = NULL;
    int     len   = 0;
    
    query = getenv("REQUEST_METHOD");
    
    if(query == NULL) {
        return NULL;
    }
    
    // GET Method
    if(strcmp(query, "GET") == 0) {
        len = strlen(getenv("QUERY_STRING"));
        query = new char[len + 1];
        check_pointer(query == NULL, return NULL);
        strcpy(query, getenv("QUERY_STRING"));
    
    // POST Method
    } else if(strcmp(query, "POST") == 0) {
        if(strstr(getenv("CONTENT_TYPE"), "www-form-urlencoded") != NULL) {
            len = atoi(getenv("CONTENT_LENGTH"));
            if(len == 0) {
                fprintf(stderr, "%s: null CONTENT_LENGTH\n", program);
                return NULL;
            }
            query = new char[len + 1];
            check_pointer(query == NULL, return NULL);
            read(0, query, len);
            //fread(query, sizeof(char), len, stdin);
            query[len] = 0;
            
        } else {
            fprintf(stderr, 
                "%s: unexpected CONTENT_TYPE: '%s'\n", 
                program, getenv("CONTENT_TYPE")
            );
            return NULL;
        }
        
    // Other: HEAD method or malformed HTTP header
    } else {
        fprintf(stderr, 
            "%s: unknown or unsupported method '%s'\n", 
            program, (query ? query : null)
        );
        return NULL;
    }
    
    // Now query is a pointer to a copy of the query string. 
    
    return query;
}


/* ***************************************************************** *
 * unescape()                                                        *
 * --------                                                          *
 * description                                                       *
 *   This function translates the url-encoded characters (in the     *
 *   form %xx, where xx is a two-digits hexadecimal number) into     *
 *   the corresponding character.                                    *
 *                                                                   *
 * ***************************************************************** */
void unescape(char * str) {
    char *  s = str;  // read pointer
    char *  p = str;  // write pointer
    
    for( ; (*p = *s) != 0; s++, p++) {
        if(*s == '%') {
            *p = hex2char(s+1);
            s += 2;
        }
    }
}


/* ***************************************************************** *
 * hex2char()                                                        *
 * --------                                                          *
 * description                                                       *
 *   This function translates a two digits hex number into the       *
 *   corresponding number, and returns the value.                    *
 *   This function is used by the unescape() function.               *
 *                                                                   *
 * ***************************************************************** */
int hex2char(char * hex) {
    int     v = 0;
    
    // Note: the following code assumes that we are running with 
    // an ASCII character set. 
    
    v  = ( (*hex >= 'A') ? (*hex | 0x20) - 'a' + 10 : *hex - '0');
    v *= 16;
    hex++;
    v += ( (*hex >= 'A') ? (*hex | 0x20) - 'a' + 10 : *hex - '0');
    
    return v;
}


/* ***************************************************************** *
 * plus2space()                                                      *
 * ----------                                                        *
 * description                                                       *
 *   This function translates the '+' characters back to real        *
 *   spaces.                                                         *
 *                                                                   *
 * ***************************************************************** */
void plus2space(char * str) {
    char *  s = str;
    
    for( ; *s != 0; s++) {
        if(*s == '+') 
            *s = ' ';
    }
}


/* ***************************************************************** *
 * space2plus()                                                      *
 * ----------                                                        *
 * description                                                       *
 *   This function translates the spaces into '+' characters.        *
 *                                                                   *
 * ***************************************************************** */
void space2plus(char * str) {
    char *  s = str;
    
    for( ; *s != 0; s++) {
        if(*s == ' ') 
            *s = '+';
    }
}


/* ***************************************************************** *
 * failure()                                                         *
 * -------                                                           *
 * description                                                       *
 *   This function is used to completly stop the program and send    *
 *   an information page explaining which error occured.             *
 *                                                                   *
 * ***************************************************************** */
void failure(char * message) {
    // HTTP Header
    printf("Status: 200 OK%s", crlf);
    printf("Content-Type: text/html%s%s", crlf, crlf);
    
    // HTML Page
    printf("<HTML>\n<HEAD>\n<TITLE>Internal Server Error</TITLE>\n</HEAD>\n\n"
           "<BODY BGColor=\"#FFFFFF\">\n"
           "<H1>Internal Server Error</H1>\n"
           "The following unrecoverable error occured while "
           "running the invoked CGI application:<BR>\n"
           "%s<BR>\n"
           "If this problem persists, please mail to "
           "<A HREF=\"webmaster@bbs.resus.univ-mrs.fr\">webmaster@bbs.resus.univ-mrs.fr</A>"
           "<BR>\n<HR>\n<I>%s v%s</I>\n"
           "</BODY>\n</HTML>", 
           message, program, version
    );
    
    fprintf(stderr, "%s: failure: %s\n", program, message);
    
    exit(1);
}


/* ***************************************************************** *
 * print_file()                                                      *
 * ----------                                                        *
 * description                                                       *
 *   This function prints the given file to the given stream.        *
 *                                                                   *
 * ***************************************************************** */
int print_file(char * filename, FILE * stream) {
    FILE *  file = NULL;
    char    buf[1025];
    int     n = 0;
    
    file = fopen(filename, "r");
    
    if(file == NULL) {
        fprintf(stderr, 
            "print_file(): fopen('%s') failed, error %d: %s\n", 
            program, filename, errno, strerror(errno)
        );
        return 0;
    }
    
    setbuf(file, NULL);
    
    while((n = fread(buf, sizeof(char), 1024, stream)) != 0) {
        fwrite(buf, sizeof(char), n, stream);
    }
    
    fclose(file);
    
    return 1;
}

